/****************************************************************************
 *
 *            Copyright (c) 2006 by CMX Systems, Inc.
 *
 * This software is copyrighted by and is the sole property of
 * CMX.  All rights, title, ownership, or other interests
 * in the software remain the property of CMX.  This
 * software may only be used in accordance with the corresponding
 * license agreement.  Any unauthorized use, duplication, transmission,
 * distribution, or disclosure of this software is expressly forbidden.
 *
 * This Copyright notice may not be removed or modified without prior
 * written consent of CMX.
 *
 * CMX reserves the right to modify this software without notice.
 *
 * CMX Systems, Inc.
 * 12276 San Jose Blvd. #511
 * Jacksonville, FL 32223
 * USA
 *
 * Tel:  (904) 880-1840
 * Fax:  (904) 880-1632
 * http: www.cmx.com
 * email: cmx@cmx.com
 *
 ***************************************************************************/
#include "../mcf5222x_reg.h"
#include "../usb-drv/usb.h"
#include "usb_cdc.h"
/****************************************************************************
 ************************** Macro definitions *******************************
 ***************************************************************************/
/* Mandatory class specific requests. */
#define CDCRQ_SEND_ENCAPSULATED_COMMAND 0x0
#define CDCRQ_GET_ENCAPSULATED_RESPONSE 0x1

/* Optional class specific requests. Windows usbser.sys depends on these. */
#define CDCRQ_SET_LINE_CODING           0x20
#define CDCRQ_GET_LINE_CODING           0x21
#define CDCRQ_SET_CONTROL_LINE_STATE    0x22

/* Optional not implemented class specific requests.
#define CDCRQ_SET_COMM_FEATURE          0x2
#define CDCRQ_GET_COMM_FEATURE          0x3
#define CDCRQ_CLEAR_COMM_FEATURE        0x4
#define CDCRQ_SEND_BREAK                0x23
*/

#define TX_EP_NO  2
#define RX_EP_NO  3

/* This macro will evaluate to an array inicializer list. It will set the
   content of a line coding structure to the defined values. */
#define FILL_LINE_CODING(bps, stop, parity, data_bits) \
  (bps) & 0xff, ((bps)>>8) & 0xff, ((bps)>>16) & 0xff, ((bps)>>24) & 0xff\
  , (hcc_u8)(stop), (hcc_u8)(parity), (hcc_u8)(data_bits)
/****************************************************************************
 ************************** Type definitions ********************************
 ***************************************************************************/

/****************************************************************************
 ************************** Function predefinitions. ************************
 ***************************************************************************/
callback_state_t got_line_coding(void);
void cdc_reset_event(void);

/****************************************************************************
 ************************** Global variables ********************************
 ***************************************************************************/
hcc_u8 new_line_coding;

/****************************************************************************
 ************************** Module variables ********************************
 ***************************************************************************/
/* Buffer long enouh to hold one USB frame. (Length equals to rx packet
   size. */
static hcc_u32 rx_buffer[32/4];
static hcc_u8 rx_length;
static hcc_u8 rx_ndx;
/* Buffer long enouh to hold one USB frame. (Length equals to tx packet
   size. */
static hcc_u8 *cur_tx_buffer;   
static hcc_u32 tx_buffer1[32/4];
static hcc_u32 tx_buffer2[32/4];
static hcc_u8 tx_ndx;

static hcc_u8 line_coding[7] = {
  FILL_LINE_CODING(19200, 0, 0, 8) /* Default is 19200 BPS and 8N1 format. */
};

/****************************************************************************
 ************************** Function definitions ****************************
 ***************************************************************************/

/*****************************************************************************
 * USB callback function. Is called by the USB driver if an USB error event
 * occurs.
 ****************************************************************************/
void usb_bus_error_event(void)
{
  /* empty */
}

/*****************************************************************************
 * USB callback function. Is called by the USB driver if an USB suspend event
 * occurs.
 ****************************************************************************/
void usb_suspend_event(void)
{
  /* empty */
}

/*****************************************************************************
 * USB callback function. Is called by the USB driver if an USB wakeup event
 * occurs.
 ****************************************************************************/
void usb_wakeup_event(void)
{
  /* empty */
}

/*****************************************************************************
 * USB callback function. Is called by the USB driver if an USB reset event
 * occurs.
 ****************************************************************************/
void usb_reset_event(void)
{
  /* empty */
}

/*****************************************************************************
 * USB callback function. Is called by the USB driver when the line coding
 * has been changed by the host.
 ****************************************************************************/
callback_state_t got_line_coding(void)
{
  new_line_coding=1;
  return(clbst_ok);
}
/*****************************************************************************
 * USB callback function. Is called by the USB driver if a non standard
 * request is received from the HOST. This callback extends the known
 * requests by masstorage related ones.
 ****************************************************************************/
callback_state_t usb_ep0_callback(void)
{
  hcc_u8 *pdata=usb_get_rx_pptr(0);
  
  callback_state_t r=clbst_error;

  /* A request to the command interface. */
  if (STP_INDEX(pdata) == CMD_IFC_INDEX)
  {
    switch(STP_REQU_TYPE(pdata))
    {
    /* Class specific in request. */
    case ((1<<7) | (1<<5) | 1):
      /* Host wants to get a descriptor */
      switch (STP_REQUEST(pdata))
      {
      case CDCRQ_GET_LINE_CODING:
        usb_send(0, (void *) 0, (void *)&line_coding, 7, STP_LENGTH(pdata));
        r=clbst_in;
        break;
      case CDCRQ_GET_ENCAPSULATED_RESPONSE:
      default:
        break;
      }
      break;
    /* Class specific out request. */
    case ((0<<7) | (1<<5) | 1):
      switch (STP_REQUEST(pdata))
      {
      case CDCRQ_SET_LINE_CODING:
        usb_receive(0, got_line_coding, (void *)&line_coding, 7, 7);
        r=clbst_out;
        break;
      case CDCRQ_SET_CONTROL_LINE_STATE:
        r=clbst_out;
        break;
      case CDCRQ_SEND_ENCAPSULATED_COMMAND:
      default:
        break;
      }
      break;
    default:
      break;
    }
  }
  return(r);
}

/*****************************************************************************
* Name:
*    cdc_getch
* In:
*    N/A
*
* Out:
*    N/A
*
* Description:
*    Get the next character from rx_buffer.
*
* Assumptions:
*    Is only called if cdc_input_ready returns true.
*****************************************************************************/
char cdc_getch(void)
{
  /* Return the next character, and advance read index. */
  return((char)((hcc_u8*)rx_buffer)[rx_ndx++]);
}

/*****************************************************************************
* Name:
*    cdc_input_ready
* In:
*    N/A
*
* Out:
*    N/A
*
* Description:
*    Will return one, if the rx_buffer contains any unread characters.
*
* Assumptions:
*    This function is called periodicaly. (Otherwise usb_receive call will
*    not bee made, and no data will ge received over the USB.)
*****************************************************************************/
int cdc_input_ready(void)
{
  /* If the receive buffer is not empty retun 1. */
  if (rx_length > rx_ndx)
  {
    return(1);
  }
  /* If the rx buffer is empty. */ 
  
  /* See if host is talking to us. */
  if (usb_get_state() != USBST_CONFIGURED)
  {
    /* If not, enter startup state. This means rx_buffer is empty, 
       and if host will be connected, the next rx transfer is
       started. */
    rx_length=1;      /* rx_length != 0 will make cdc_input_ready to start
                         the next rx transfer on USB. */
    rx_ndx=rx_length; /* Enter rx buffer is empty state. */        
    return(0);
  }
  
  /* if rx_length != 0 then the buffer has been empty and we need to start
     the next rx transfer on USB. */
  if (rx_length > 0)
  {
    rx_length = 0;  /* Do this to avoid getting here while the next frame
                       is not received and processed by the main app. */
    rx_ndx=0;       /* Set rx buffer state to unread. */
    /* We have no information how much data the host is willing to send, 
       and thus we don't know when the started transmission will end.
       The usb driver will treat a trensmission ended if the number of 
       needed bytes has been received, or if a short packet has been 
       received. So we ask the driver to receive data that fits exactly 
       one packet. This way after the first frame the transfer will end.
       (If the host sends less data than a packet, then it will end any way,
       and if the host wants to send more data then a packet can hold, then 
       we will make an usb_receive call for each packet.
       If we wold try to receive let's syat 64k, and the host has not that 
       much data, then we would wait till the end of time for data. */
    usb_receive(RX_EP_NO, (void *) 0, (void *) rx_buffer, sizeof(rx_buffer), sizeof(rx_buffer));
    return(0);
  }  
   
  /* If we get here, then the rx buffer is empty, and a frame reception is
     started. So we check for the status of the reception. */
  if (!usb_ep_is_busy(RX_EP_NO))
  {
    /* Read out error status of endpoint. The reception may be aborted due
       to a status change on the USB (disconnect, sleep, etc..) or due to
       an error (CRC, bit stuffing, etc...). In both case wee need to restart
       reception if possible.  */
    switch(usb_ep_error(RX_EP_NO))
    {
    case USBEPERR_NONE: /* Reception finished with no error. */
      /* Read out number of received bytes. This will make us to return 
         received characters for the next call. Note: the transfer may 
         contain 0 characters, so we return the first character only when
         the next call is made. */
      rx_length=(hcc_u8)usb_get_done(RX_EP_NO);
      break;
    case USBEPERR_PROTOCOL_ERROR:
      /* If we have an error, then we restart the reception. */
      usb_receive(RX_EP_NO, (void *) 0, (void *) rx_buffer, sizeof(rx_buffer), sizeof(rx_buffer));
      break;
    case USBEPERR_HOST_ABORT:
      /* USB disconnected or in low power mode. 
         Do nothing because this is handled by the second if above. */
      break;
    case USBEPERR_USER_ABORT:
    case USBEPERR_TO_MANY_DATA:    
    default:
      /* Upps! unexpected error. Stop here. */
      while(1)
      ;
    }
  }
  return(0);
}

/*****************************************************************************
* Name:
*    cdc_putch
* In:
*    0: output is busy, character dropped
*    1: character buffered and will be sent later.
* Out:
*    N/A
*
* Description:
*    Put one character into tx_buffer.
*
* Assumptions:
*    --
*****************************************************************************/
int cdc_putch(hcc_u8 c)
{
  int r=0;
  /* Store character into current buffer. */
  if (tx_ndx < sizeof(tx_buffer1))
  {
    r=1;
    cur_tx_buffer[tx_ndx++]=(hcc_u8)c;
  }
  
  /* Hndle USB. See if we need to send something. */
  if (tx_ndx != 0)
  {
    /* Host properly connected us, and the TX endpoint is not busy */
    if (usb_get_state() == USBST_CONFIGURED)
    {
      if (!usb_ep_is_busy(TX_EP_NO))
      {
        /* Check the status of the next transfer. */   
        switch (usb_ep_error(TX_EP_NO))
        {
        case USBEPERR_NONE: /* Finished with no error. */
          /* Start sending next chunk. */
          usb_send(TX_EP_NO, (void *) 0, (void *) cur_tx_buffer, tx_ndx, tx_ndx);
          /* Switch buffer. */
          tx_ndx=0;
          cur_tx_buffer = (hcc_u8*)((cur_tx_buffer == (hcc_u8*)tx_buffer1) ? tx_buffer2 : tx_buffer1);
          break;
        case USBEPERR_PROTOCOL_ERROR:
        case USBEPERR_HOST_ABORT:
          /* It is possible to resend the previous buffer here. */
          break;
        case USBEPERR_USER_ABORT:
        case USBEPERR_TO_MANY_DATA:      
        default:
          /* Upps! unexpected error. Stop here. */
          while(1)
          ;          
        }
      }
    }    
  }
  return(r);
}

/*****************************************************************************
 *
 ****************************************************************************/
int cdc_line_coding_changed(void)
{
  if(new_line_coding)
  {
    new_line_coding=0;
    return(1);
  }
  return(0);
}

/*****************************************************************************
 *
 ****************************************************************************/
void cdc_get_line_coding(line_coding_t *l)
{
  l->bps=RD_LE32(line_coding);
  l->nstp=line_coding[4];
  l->parity=line_coding[5];
  l->ndata=line_coding[6];
}

/*****************************************************************************
 *
 ****************************************************************************/
void cdc_init(void)
{
  cur_tx_buffer=(hcc_u8*)tx_buffer1;   
  tx_ndx=0; 
}
/****************************** END OF FILE **********************************/
